/*
 * double linked list / btree (well, just DLL at the moment)
 *
 * node->prev == NULL at the head of the list
 * node->next == NULL at the tail of the list
 */

#include <llist.h>
#include <log.h>
#include <assert.h>

/*
 *    function: node *ll_append()
 * description: insert item at the end of the list
 *     returns: pointer to the new node, NULL on malloc failure
 */
node *
ll_append(l, data)
    llist	*l;
    void	*data;
{
    node	*new;

    if (debug & DEBUG_CALL)
	debug_msg("ll_append(%x, %x)\n", l, data);

    if ((new = (node *) malloc(sizeof(node))) == NULL)
	return(NULL);
    bzero(new, sizeof(node));

    new->data = data;

    /* stick it on the end */
    if (l->head == NULL) {
	l->head = new;
	l->butt = new;
    } else {
	l->butt->next = new;
	new->prev = l->butt;
	l->butt = new;
    }

    return(new);
}

/*
 *    function: void ll_clear()
 * description: clear a llist
 *     returns:
 */
void
ll_clear(llist **l)
{
    node	*current;

    if (debug & DEBUG_CALL)
	debug_msg("ll_clear(%x)\n", l);

    if (l == NULL)
	return;

    current = (*l)->head;

    while (current) {
	(*l)->head = current->next;
	if (current->data)
	    free(current->data);
	if (current->key_t == LL_KEY_STR && current->key.str != NULL)
	    free(current->key.str);
	free(current);
	current = (*l)->head;
    }

    free(*l);
    *l = NULL;

    return;
}

/*
 *    function: int ll_count()
 * description: count number of nodes on a LL
 *     returns: number of nodes, -1 on error
 */
int
ll_count(l)
    llist	*l;
{
    int		cnt = 0;
    node	*n;

    if (debug & DEBUG_CALL)
	debug_msg("ll_count(%x)\n", l);

    if (l == NULL)
	return(-1);

    n = l->head;
    while (n != NULL) {
	cnt++;
	n = n->next;
    }

    return(cnt);
}

/*
 *    function: llist *ll_create()
 * description: create a new empty llist
 *     returns: pointer to a new llist struct, else NULL due to malloc error -
 *		see errno
 */
llist *
ll_create()
{
    llist	*l;

    if (debug & DEBUG_CALL)
	debug_msg("ll_create()\n");

    if ((l = (llist *) malloc(sizeof(llist))) == NULL)
	return(NULL);

    bzero(l, sizeof(llist));

    return(l);
}

/*
 *    function: node *ll_delete()
 * description: delete an entry in a linked list
 *     returns: node * which the caller must free(), or NULL on error
 */
node *
ll_delete(l, n)
    llist	*l;
    node	*n;
{
    if (debug & DEBUG_CALL)
	debug_msg("ll_delete(%x, %x)\n", l, n);

    if (l == NULL || n == NULL)
	return(NULL);

    if (n->prev == NULL) {
	/* head node */
	l->head = n->next;
    } else {
	/* adjust the prev->next */
	n->prev->next = n->next;
    }

    if (n->next == NULL) {
	/* tail node, prev may be a node or NULL */
	l->butt = n->prev;
    } else {
	/* adjust the next->prev */
	n->next->prev = n->prev;
    }

    /* clean-up returning node */
    n->prev = n->next = NULL;

    return(n);
}

/*
 *    function: node *ll_insert_n()
 * description: insert a node in a linked list before given node *i.  if *i
 *		is NULL, insert at the tail.
 *     returns: node *i, or NULL on error
 */
node *
ll_insert_n(l, i, n)
    llist	*l;
    node	*i,
		*n;
{
    if (debug & DEBUG_CALL)
	debug_msg("ll_insert(%x, %x, %x)\n", l, i, n);

    if (l == NULL || n == NULL)
	return(NULL);

    /* if NULL insertion node, just append to the LL */
    if (i == NULL) {
	if (l->head == NULL) {
	    l->head = n;
	    l->butt = n;
	} else {
	    l->butt->next = n;
	    n->prev = l->butt;
	    l->butt = n;
	}
    } if (i->prev == NULL) {
	/* i is head of the LL */
	n->next = l->head;
	if (l->head != NULL)
	    l->head->prev = n;
	l->head = n;
    } else {
	n->prev = i->prev;
	i->prev = n;
	n->prev->next = n;
	n->next = i;
    }

    return(i);
}

/*
 *    function: node *ll_next()
 * description: next node on llist
 *     returns: pointer to next node on llist, else NULL
 */
node *
ll_next(l, n)
    llist	*l;
    node	*n;
{
    if (debug & DEBUG_CALL)
	debug_msg("ll_next(%x)\n", n);

    if (n == NULL)
	return(l->head);

    return(n->next);
}

/*
 *    function: node *ll_node_n()
 * description: find node N of a LL
 *     returns: pointer to node N on llist starting at 0, else NULL
 */
node *
ll_node_n(l, n)
    llist	*l;
    int		n;
{
    node	*x;
    int		cnt = 0;

    if (debug & DEBUG_CALL)
	debug_msg("ll_node_n(%x, %d)\n", l, n);

    if (l == NULL || n < 0)
	return(NULL);

    x = l->head;
    while (x != NULL && cnt != n) {
	x = x->next;
	cnt++;
    }

    return(x);
}

/*
 *    function: node *ll_pop()
 * description: pop first node on the list
 *     returns: pointer to first node on llist, else NULL
 */
node *
ll_pop(l)
    llist	*l;
{
    node	*first;

    if (debug & DEBUG_CALL)
	debug_msg("ll_pop(%x)\n", l);

    if (l == NULL || l->head == NULL)
	return(NULL);

    /* adjust the LL head and clean-up returning node */
    first = l->head;
    l->head = first->next;
    if (l->head != NULL)
	l->head->prev = NULL;
    else
	l->butt = NULL;
    first->next = NULL;

    return(first);
}

/*
 *    function: node *ll_prev()
 * description: previous node on llist
 *     returns: pointer to previous node on DLL, else NULL
 */
node *
ll_prev(n)
    node	*n;
{
    if (debug & DEBUG_CALL)
	debug_msg("ll_prev(%x)\n", n);

    return(n->prev);
}

/*
 *    function: node *ll_push()
 * description: insert item at the beginning of the list (ie: push)
 *     returns: pointer to the new node
 */
node *
ll_push(l, data)
    llist	*l;
    void	*data;
{
    node	*new;

    if (debug & DEBUG_CALL)
	debug_msg("ll_push(%x, %x)\n", l, data);

    assert(l != NULL);

    if ((new = (node *) malloc(sizeof(node))) == NULL)
	return(NULL);
    bzero(new, sizeof(node));

    new->data = data;

    /* push */
    new->next = l->head;
    if (l->head != NULL)
	l->head->prev = new;
    l->head = new;

    if (l->butt == NULL)
	l->butt = new;

    return(new);
}

/*
 *    function: llist *ll_shift()
 * description: remove item from end of the list
 *     returns: pointer to the node
 */
node *
ll_shift(l)
    llist	*l;
{
    node	*last;

    if (debug & DEBUG_CALL)
	debug_msg("ll_shift(%x)\n", l);

    if (l == NULL || l->head == NULL)
	return(NULL);

    /* end of the list */
    last = l->butt;

    if (last->prev == NULL) {
	/* node is both beginning and end of list, so list is now empty */
	l->head = NULL;
	l->butt = NULL;
    } else {
	/* else, adjust the tail node and clean-up the returning node */
	last->prev->next = NULL;
	l->butt = last->prev;
	last->prev = NULL;
    }

    return(last);
}

/*
 *    function: int ll_sort()
 * description: sort nodes of the LL based on the node keys.  duplicate keys
 *		are inserted before dupes.
 *     returns: non-zero on error
 */
int
ll_sort(l)
    llist	*l;
{
    int		val = 0;
    node	*n,
		*head,
		*last;

    if (debug & DEBUG_CALL)
	debug_msg("ll_sort(%x)\n", l);

    if (l == NULL || l->head == NULL)
	return(-1);

    last = l->head;
    n = last->next;

	/* HEAS: this should handle a user-supplied compare function */
    /*
     * for each node following the *last node, ll_delete() it and traverse
     * the list from the head up to *last.  if inserted after *last, n
     * becomes the new *last.
     */
    while (n != NULL) {
	/* compare *n to *last, if greater than, leave and make it *last */
	if (n->key_t == LL_KEY_INT) {
	    if (n->key.i >= last->key.i) {
		last = n;
		goto NEXT;
	    }
	} else if (n->key_t == LL_KEY_UINT) {
	    if (n->key.ui >= last->key.ui) {
		last = n;
		goto NEXT;
	    }
	} else { /* LL_KEY_STR */
	    if (strcmp(n->key.str, last->key.str) >= 0) {
		last = n;
		goto NEXT;
	    }
	}

	/* remove the node from the list and start looking for the insertion
	 * point from the LL head.
	 */
	ll_delete(l, n);
	head = l->head;
	while (head != NULL && head != last) {
	    if (n->key_t == LL_KEY_INT) {
		if (n->key.i <= head->key.i)
		    break;
	    } else if (n->key_t == LL_KEY_UINT) {
		if (n->key.ui <= head->key.ui)
		    break;
	    } else { /* LL_KEY_STR */
		if (strcmp(n->key.str, head->key.str) <= 0)
		    break;
	    }

	    head = head->next;
	}
	/* head is now at our insertion point */
	ll_insert_n(l, head, n);

NEXT:	n = last->next;
    }

    return(val);
}
